# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable

import pydiverse.common as pdc

from ..optional_dependency import ColExpr, dy
from ._base import Column


class Any(Column):
    """A column that can contain any type."""

    def __init__(
        self,
        *,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        super().__init__(nullable=True, primary_key=False, check=check, alias=alias)

    def dtype(self) -> pdc.Dtype:
        raise NotImplementedError(
            "The Type Any is intentionally not implemented in pydiverse libraries."
        )

    def to_dataframely(self):
        return dy.Any(check=self.check, alias=self.alias)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

from pydiverse.colspec import Collection, ColSpec, Column, Filter, Rule, RulePolars


def create_colspec(
    name: str,
    columns: dict[str, Column],
    rules: dict[str, Rule | RulePolars] | None = None,
) -> type[ColSpec]:
    """Dynamically create a new column specification with the provided name.

    Args:
        name: The name of the column specification.
        columns: The columns to set on the column specification. When properly defining
            the column specification, this would be the annotations that define the
            column types.
        rules: The custom non-column-specific validation rules. When properly defining
            the column specification, this would be the functions annotated with
            ``@dy.rule``.

    Returns:
        The dynamically created column specification.
    """
    return type(name, (ColSpec,), {**columns, **(rules or {})})


def create_collection(
    name: str,
    colspecs: dict[str, type[ColSpec]],
    filters: dict[str, Filter] | None = None,
) -> type[Collection]:
    return create_collection_raw(
        name,
        annotations={
            name: colspec  # type: ignore
            for name, colspec in colspecs.items()
        },
        filters=filters,
    )


def create_collection_raw(
    name: str,
    annotations: dict[str, Any],
    filters: dict[str, Filter] | None = None,
) -> type[Collection]:
    return type(
        name,
        (Collection,),
        {
            "__annotations__": annotations,
            **(filters or {}),
        },
    )



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable

import pydiverse.common as pdc

from ..optional_dependency import ColExpr, pl
from ._base import Column


class Struct(Column):
    """A struct column."""

    def __init__(
        self,
        inner: dict[str, Column],
        *,
        nullable: bool = True,
        primary_key: bool = False,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            inner: The dictionary of struct fields. Struct fields may have
                ``primary_key=True`` set but this setting only takes effect if the
                struct is nested inside a list. In this case, the list items must be
                unique wrt. the struct fields that have ``primary_key=True`` set.
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        super().__init__(
            nullable=nullable, primary_key=primary_key, check=check, alias=alias
        )
        self.inner = inner

    def dtype(self) -> pdc.Dtype:
        raise NotImplementedError(
            "Struct column type is not yet implemented in pydiverse libraries."
        )

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        raise NotImplementedError(
            "Struct column type is not yet implemented in pydiverse libraries."
        )
        inner_rules = {
            f"inner_{name}_{rule_name}": (
                pl.when(expr.is_null()).then(pl.lit(True)).otherwise(inner_expr)
            )
            for name, col in self.inner.items()
            for rule_name, inner_expr in col.validation_rules(
                expr.struct.field(name)
            ).items()
        }
        return {
            **super().validation_rules(expr),
            **inner_rules,
        }



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.common as pdc

from ._base import Column

# ------------------------------------------------------------------------------------ #


class Bool(Column):
    """A column of booleans."""

    def dtype(self) -> pdc.Bool:
        return pdc.Bool()



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import C, dy, pdt, pl


class AliasColSpec(cs.ColSpec):
    a = cs.Int64(alias="hello world: col with space!")


def test_column_names():
    assert AliasColSpec.column_names() == ["hello world: col with space!"]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_validation():
    df = pl.DataFrame({"hello world: col with space!": [1, 2]})
    assert AliasColSpec.is_valid_polars(df)
    tbl = pdt.Table(df)
    assert AliasColSpec.is_valid(tbl)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_create_empty():
    df = AliasColSpec.create_empty_polars()
    assert AliasColSpec.is_valid_polars(df)
    tbl = pdt.Table(df)
    assert AliasColSpec.is_valid(tbl)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from .. import (
    Any,
    ColSpec,
    Date,
    Datetime,
    Decimal,
    Enum,
    Float32,
    Int64,
    List,
    Struct,
)


class MyImportedBaseColSpec(ColSpec):
    a = Int64()


class MyImportedColSpec(MyImportedBaseColSpec):
    b = Float32()
    c = Enum(["a", "b", "c"])
    d = Struct({"a": Int64(), "b": Struct({"c": Enum(["a", "b"])})})
    e = List(Struct({"a": Int64()}))
    f = Datetime()
    g = Date()
    h = Any()
    some_decimal = Decimal(12, 8)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import random

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.exc import DtypeValidationError, ValidationError
from pydiverse.colspec.optional_dependency import (
    C,
    DataTypeClass,
    assert_frame_equal,
    pdt,
    pl,
    sa,
)
from pydiverse.colspec.testing.assert_equal import assert_table_equal

if sa.Column is not None:
    engine = sa.create_engine("duckdb:///:memory:")
else:
    engine = None


def sql_table(df: pl.DataFrame, *, name: str) -> pdt.Table:
    df.write_database(name, engine, if_table_exists="replace")
    return pdt.Table(name, pdt.SqlAlchemy(engine))


class MyColSpec(cs.ColSpec):
    a = cs.Int64(primary_key=True)
    b = cs.String(max_length=3)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    ("schema", "expected_columns"),
    [
        ({"a": pl.Int64, "c": pl.String}, None),
        ({"a": pl.Int64, "c": pl.String}, None),
        ({"a": pl.Int64, "b": pl.String, "c": pl.String}, ["a", "b"]),
    ],
)
def test_filter_extra_columns(
    schema: dict[str, DataTypeClass], expected_columns: list[str] | None
):
    tbl = sql_table(pl.DataFrame(schema=schema), name="tbl")
    try:
        filtered, _ = MyColSpec.filter(tbl)
        filtered >>= pdt.export(pdt.Polars)
        assert expected_columns is not None
        assert set(filtered.columns) == set(expected_columns)
    except ValidationError:
        assert expected_columns is None


@pytest.mark.parametrize(
    ("schema", "cast", "success"),
    [
        ({"a": pl.Int64, "b": pl.Int64}, False, False),
        ({"a": pl.String, "b": pl.String}, True, True),
    ],
)
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_dtypes(schema: dict[str, DataTypeClass], cast: bool, success: bool):
    tbl = sql_table(pl.DataFrame(schema=schema), name="tbl")
    try:
        MyColSpec.filter(tbl, cast=cast)
        assert success
    except DtypeValidationError:
        assert not success


@pytest.mark.parametrize(
    ("data_a", "data_b", "failure_mask", "counts", "cooccurrence_counts"),
    [
        ([1, 2, 3], ["foo", "bar", None], [True, True, True], {}, {}),
        (
            [1, 2, 3],
            ["foo", "bar", "foobar"],
            [True, True, False],
            {"b|max_length": 1},
            {frozenset({"b|max_length"}): 1},
        ),
        (
            [1, 2, 2],
            ["foo", "bar", "foobar"],
            [True, False, False],
            {"b|max_length": 1, "_primary_key_": 2},
            {
                frozenset({"b|max_length", "_primary_key_"}): 1,
                frozenset({"_primary_key_"}): 1,
            },
        ),
    ],
)
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_failure(
    data_a: list[int],
    data_b: list[str | None],
    failure_mask: list[bool],
    counts: dict[str, int],
    cooccurrence_counts: dict[frozenset[str], int],
):
    tbl = sql_table(pl.DataFrame({"a": data_a, "b": data_b}), name="tbl")
    tbl_valid, failures = MyColSpec.filter(tbl)
    assert isinstance(tbl_valid, pdt.Table)
    assert_frame_equal(
        (tbl >> pdt.export(pdt.Polars)).filter(pl.Series(failure_mask)),
        tbl_valid >> pdt.export(pdt.Polars),
        check_row_order=False,
    )
    assert len(failures) == (len(failure_mask) - sum(failure_mask))
    assert failures.counts() == counts


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_no_rules():
    class TestColSpec(cs.ColSpec):
        a = cs.Int64(nullable=True)

    tbl = sql_table(pl.DataFrame({"a": [1, 2, 3]}), name="tbl")
    df_valid, failures = TestColSpec.filter(tbl)
    assert isinstance(df_valid, pdt.Table)
    assert_table_equal(tbl, df_valid)
    assert len(failures) == 0
    assert failures.counts() == {}


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_with_rule_all_valid():
    class TestColSpec(cs.ColSpec):
        a = cs.String(min_length=3)

    tbl = sql_table(pl.DataFrame({"a": ["foo", "foobar"]}), name="tbl")
    tbl_valid, failures = TestColSpec.filter(tbl)
    assert isinstance(tbl_valid, pdt.Table)
    assert_table_equal(tbl, tbl_valid)
    assert len(failures) == 0
    assert failures.counts() == {}


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_cast():
    data = {
        # validation: [true, true, false, false, false, false]
        "a": ["1", "2", "foo", None, "123x", "9223372036854775808"],
        # validation: [true, false, true, true, false, true]
        "b": [20, 2000, None, 30, 3000, 50],
    }
    tbl = sql_table(pl.DataFrame(data), name="tbl")
    tbl_valid, failures = MyColSpec.filter(tbl, cast=True)
    assert isinstance(tbl_valid, pdt.Table)
    assert [col.name for col in tbl_valid] == MyColSpec.column_names()
    assert len(failures) == 5
    assert failures.counts() == {
        "a|dtype": 3,
        "a|nullability": 1,
        "b|max_length": 1,
    }
    # assert failures.cooccurrence_counts() == {
    #     frozenset({"a|nullability", "primary_key"}): 1,
    #     frozenset({"b|max_length"}): 1,
    #     frozenset({"a|dtype"}): 3,
    # }


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_nondeterministic_tbl():
    n = 10_000
    tbl = sql_table(
        pl.DataFrame(
            {
                "a": range(n),
                "b": [random.choice(["foo", "foobar"]) for _ in range(n)],
            }
        ).select(pl.all().shuffle()),
        name="tbl",
    )

    filtered, _ = MyColSpec.filter(tbl)
    assert (
        filtered
        >> pdt.group_by(filtered.b)
        >> pdt.summarize()
        >> pdt.alias()
        >> pdt.summarize(n_unique=pdt.count())
        >> pdt.export(pdt.Scalar)
    ) == 1



# Copyright (c) QuantCo and pydiverse contributors 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

from abc import abstractmethod
from collections.abc import Callable, Sequence

import pydiverse.common as pdc

from ..optional_dependency import ColExpr
from ._base import Column
from ._mixins import IsInMixin, OrdinalMixin
from ._utils import classproperty


class _BaseInteger(IsInMixin[int], OrdinalMixin[int], Column):
    def __init__(
        self,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        min: int | None = None,  # noqa: A002
        min_exclusive: int | None = None,
        max: int | None = None,  # noqa: A002
        max_exclusive: int | None = None,
        is_in: Sequence[int] | None = None,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            min: The minimum value for integers in this column (inclusive).
            min_exclusive: Like ``min`` but exclusive. May not be specified if ``min``
                is specified and vice versa.
            max: The maximum value for integers in this column (inclusive).
            max_exclusive: Like ``max`` but exclusive. May not be specified if ``max``
                is specified and vice versa.
            is_in: A (non-contiguous) list of integers indicating valid values in this
                column. If specified, both ``min`` and ``max`` must not bet set.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        if min is not None and min < self.min_value:
            raise ValueError("`min` is too small for the data type.")
        if max is not None and max > self.max_value:
            raise ValueError("`max` is too big for the data type.")
        if is_in is not None and (min is not None or max is not None):
            raise ValueError(
                "`is_in` may only be specified if `min` and `max` are unspecified."
            )

        super().__init__(
            nullable=nullable,
            primary_key=primary_key,
            min=min,
            min_exclusive=min_exclusive,
            max=max,
            max_exclusive=max_exclusive,
            is_in=is_in,
            check=check,
            alias=alias,
        )

    @classproperty
    @abstractmethod
    def num_bytes(self) -> int:
        """Number of bytes that the column type consumes."""

    @classproperty
    @abstractmethod
    def is_unsigned(self) -> bool:
        """Whether the column type is unsigned."""

    @classproperty
    def max_value(self) -> int:
        """Maximum value of the column's type."""
        return (
            2 ** (self.num_bytes * 8) - 1
            if self.is_unsigned
            else 2 ** (self.num_bytes * 8 - 1) - 1
        )

    @classproperty
    def min_value(self) -> int:
        """Minimum value of the column's type."""
        return 0 if self.is_unsigned else -(2 ** (self.num_bytes * 8 - 1))


# ------------------------------------------------------------------------------------ #


class Integer(_BaseInteger):
    """A column of integers (with any number of bytes)."""

    def dtype(self) -> pdc.Int:
        return pdc.Int()

    @classproperty
    def num_bytes(self) -> int:
        return 8

    @classproperty
    def is_unsigned(self) -> bool:
        return False


class Int8(_BaseInteger):
    """A column of int8 values."""

    def dtype(self) -> pdc.Int8:
        return pdc.Int8()

    @classproperty
    def num_bytes(self) -> int:
        return 1

    @classproperty
    def is_unsigned(self) -> bool:
        return False


class Int16(_BaseInteger):
    """A column of int16 values."""

    def dtype(self) -> pdc.Int16:
        return pdc.Int16()

    @classproperty
    def num_bytes(self) -> int:
        return 2

    @classproperty
    def is_unsigned(self) -> bool:
        return False


class Int32(_BaseInteger):
    """A column of int32 values."""

    def dtype(self) -> pdc.Int32:
        return pdc.Int32()

    @classproperty
    def num_bytes(self) -> int:
        return 4

    @classproperty
    def is_unsigned(self) -> bool:
        return False


class Int64(_BaseInteger):
    """A column of int64 values."""

    def dtype(self) -> pdc.Int64:
        return pdc.Int64()

    @classproperty
    def num_bytes(self) -> int:
        return 8

    @classproperty
    def is_unsigned(self) -> bool:
        return False


class UInt8(_BaseInteger):
    """A column of uint8 values."""

    def dtype(self) -> pdc.UInt8:
        return pdc.UInt8()

    @classproperty
    def num_bytes(self) -> int:
        return 1

    @classproperty
    def is_unsigned(self) -> bool:
        return True


class UInt16(_BaseInteger):
    """A column of uint16 values."""

    def dtype(self) -> pdc.UInt16:
        return pdc.UInt16()

    @classproperty
    def num_bytes(self) -> int:
        return 2

    @classproperty
    def is_unsigned(self) -> bool:
        return True


class UInt32(_BaseInteger):
    """A column of uint32 values."""

    def dtype(self) -> pdc.UInt32:
        return pdc.UInt32()

    @classproperty
    def num_bytes(self) -> int:
        return 4

    @classproperty
    def is_unsigned(self) -> bool:
        return True


class UInt64(_BaseInteger):
    """A column of uint64 values."""

    def dtype(self) -> pdc.UInt64:
        return pdc.UInt64()

    @classproperty
    def num_bytes(self) -> int:
        return 8

    @classproperty
    def is_unsigned(self) -> bool:
        return True



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import dy, pl
from pydiverse.colspec.testing.factory import create_colspec


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("dy_enum", "pl_dtype", "valid"),
    [
        (cs.Enum(["x", "y"]), pl.Enum(["x", "y"]), True),
        (cs.Enum(["y", "x"]), pl.Enum(["x", "y"]), False),
        (cs.Enum(["x"]), pl.Enum(["x", "y"]), False),
        (cs.Enum(["x", "y", "z"]), pl.Enum(["x", "y"]), False),
        (cs.Enum(["x", "y"]), pl.String(), False),
    ],
)
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_valid(
    df_type: type[pl.DataFrame] | type[pl.LazyFrame],
    dy_enum: cs.Enum,
    pl_dtype: pl.Enum,
    valid: bool,
):
    schema = create_colspec("test", {"a": dy_enum})
    df = df_type({"a": ["x", "y", "x", "x"]}).cast(pl_dtype)
    assert schema.is_valid_polars(df) == valid


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("enum", [cs.Enum(["x", "y"]), cs.Enum(["y", "x"])])
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
@pytest.mark.parametrize(
    ("data", "valid"),
    [
        ({"a": ["x", "y", "x", "x"]}, True),
        ({"a": ["x", "y", "x", "x"]}, True),
        ({"a": ["x", "y", "z"]}, False),
        ({"a": ["x", "y", "z"]}, False),
    ],
)
def test_valid_cast(
    enum: cs.Enum,
    data: Any,
    valid: bool,
    df_type: type[pl.DataFrame] | type[pl.LazyFrame],
):
    schema = create_colspec("test", {"a": enum})
    df = df_type(data)
    assert schema.is_valid_polars(df, cast=True) == valid



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import dy, pl


class MyColSpec(cs.ColSpec):
    a = cs.Int64()
    b = cs.String()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_create_empty():
    df = MyColSpec.create_empty_polars()
    assert df.columns == ["a", "b"]
    assert df.dtypes == [pl.Int64, pl.String]
    assert len(df) == 0



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import re

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec._validation import DtypeCasting, validate_dtypes
from pydiverse.colspec.exc import DtypeValidationError
from pydiverse.colspec.optional_dependency import (
    C,
    assert_frame_equal,
    dy,
    pdt,
    pl,
    plexc,
)


@pytest.mark.skipif(C is None, reason="pydiverse.transform is required for this test")
@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("actual", "expected", "casting"),
    [
        ({"a": pl.Int64()}, {"a": cs.Int64()}, "none"),
        ({"a": pl.Int32()}, {"a": cs.Int64()}, "lenient"),
        ({"a": pl.Int32()}, {"a": cs.Int64()}, "strict"),
        (
            {"a": pl.Int32(), "b": pl.String()},
            {"a": cs.Int64(), "b": cs.UInt8()},
            "strict",
        ),
    ],
)
def test_success(
    actual: dict[str, pl.DataType],
    expected: dict[str, cs.Column],
    casting: DtypeCasting,
):
    df = pl.DataFrame(schema=actual)
    tbl = pdt.Table(df.lazy())
    lf = validate_dtypes(tbl, expected=expected, casting=casting) >> pdt.export(
        pdt.Polars
    )
    schema = lf.collect_schema()
    for key, col in expected.items():
        assert col.validate_dtype_polars(schema[key])


@pytest.mark.skipif(C is None, reason="pydiverse.transform is required for this test")
@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("actual", "expected", "error", "fail_columns"),
    [
        (
            {"a": pl.Int32()},
            {"a": cs.Int64()},
            r"1 columns have an invalid dtype.*\n.*got dtype 'Int32'",
            {"a"},
        ),
        (
            {"a": pl.Int32(), "b": pl.String()},
            {"a": cs.Int64(), "b": cs.UInt8()},
            r"2 columns have an invalid dtype",
            {"a", "b"},
        ),
    ],
)
def test_failure(
    actual: dict[str, pl.DataType],
    expected: dict[str, cs.Column],
    error: str,
    fail_columns: set[str],
):
    df = pl.DataFrame(schema=actual)
    tbl = pdt.Table(df.lazy())
    try:
        validate_dtypes(tbl, expected=expected, casting="none")
        raise AssertionError()  # above should raise
    except DtypeValidationError as exc:
        assert set(exc.errors.keys()) == fail_columns
        assert re.match(error, str(exc))


@pytest.mark.skipif(C is None, reason="pydiverse.transform is required for this test")
@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_lenient_casting():
    lf = pl.LazyFrame(
        {"a": [1, 2, 3], "b": ["foo", "12", "1313"]},
        schema={"a": pl.Int64(), "b": pl.String()},
    )
    tbl = pdt.Table(lf)
    actual = validate_dtypes(
        tbl,
        expected={"a": cs.UInt8(), "b": cs.UInt8()},
        casting="lenient",
    ) >> pdt.export(pdt.Polars(lazy=True))
    expected = pl.LazyFrame(
        {"a": [1, 2, 3], "b": [None, 12, None]},
        schema={"a": pl.UInt8(), "b": pl.UInt8()},
    )
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(C is None, reason="pydiverse.transform is required for this test")
@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_strict_casting():
    lf = pl.LazyFrame(
        {"a": [1, 2, 3], "b": ["foo", "12", "1313"]},
        schema={"a": pl.Int64(), "b": pl.String()},
    )
    tbl = pdt.Table(lf)
    lf_valid = validate_dtypes(
        tbl,
        expected={"a": cs.UInt8(), "b": cs.UInt8()},
        casting="strict",
    ) >> pdt.export(pdt.Polars(lazy=True))
    with pytest.raises(
        plexc.InvalidOperationError, match=r'for 2 out of 3 values: \["foo", "1313"\]'
    ):
        lf_valid.collect()



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.common as pdc
from pydiverse.colspec.exc import (
    DtypeValidationError,
    RuleValidationError,
    ValidationError,
)


def test_validation_error_str():
    message = "validation failed"
    exc = ValidationError(message)
    assert str(exc) == message


def test_dtype_validation_error_str():
    exc = DtypeValidationError(
        errors={"a": (pdc.Int64(), pdc.String()), "b": (pdc.Bool(), pdc.String())}
    )
    assert str(exc).split("\n") == [
        "2 columns have an invalid dtype:",
        " - 'a': got dtype 'Int64' but expected 'String'",
        " - 'b': got dtype 'Bool' but expected 'String'",
    ]


def test_rule_validation_error_str():
    exc = RuleValidationError(
        {
            "b|max_length": 1500,
            "a|nullability": 2,
            "primary_key": 2000,
            "a|min_length": 5,
        },
    )
    assert str(exc).split("\n") == [
        "4 rules failed validation:",
        " - 'primary_key' failed validation for 2,000 rows",
        " * Column 'a' failed validation for 2 rules:",
        "   - 'min_length' failed for 5 rows",
        "   - 'nullability' failed for 2 rows",
        " * Column 'b' failed validation for 1 rules:",
        "   - 'max_length' failed for 1,500 rows",
    ]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

try:
    import pydiverse.transform as pdt

    ColExpr = pdt.ColExpr
except ImportError:
    pdt = None

    # Only serves as type hint
    class ColExpr:
        str = None
        dt = None
        dur = None


from ._base import Column
from .any import Any
from .bool import Bool
from .datetime import Date, Datetime, Duration, Time
from .decimal import Decimal
from .enum import Enum
from .float import Float, Float32, Float64
from .integer import Int8, Int16, Int32, Int64, Integer, UInt8, UInt16, UInt32, UInt64
from .list import List
from .string import String
from .struct import Struct

__all__ = [
    "Column",
    "Any",
    "Bool",
    "Date",
    "Datetime",
    "Decimal",
    "Duration",
    "Enum",
    "Time",
    "Float",
    "Float32",
    "Float64",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Integer",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "String",
    "List",
    "Struct",
]



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import random

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.exc import DtypeValidationError, ValidationError
from pydiverse.colspec.optional_dependency import (
    C,
    DataTypeClass,
    assert_frame_equal,
    dy,
    pdt,
    pl,
    validation_mask,
)


class MyColSpec(cs.ColSpec):
    a = cs.Int64(primary_key=True)
    b = cs.String(max_length=3)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    ("schema", "expected_columns"),
    [
        ({"a": pl.Int64, "c": pl.String}, None),
        ({"a": pl.Int64, "c": pl.String}, None),
        ({"a": pl.Int64, "b": pl.String, "c": pl.String}, ["a", "b"]),
    ],
)
def test_filter_extra_columns(
    schema: dict[str, DataTypeClass], expected_columns: list[str] | None
):
    df = pl.DataFrame(schema=schema)
    try:
        filtered, _ = MyColSpec.filter_polars(df)
        assert expected_columns is not None
        assert set(filtered.columns) == set(expected_columns)
    except ValidationError:
        assert expected_columns is None
    except Exception as e:
        raise AssertionError() from e
    tbl = pdt.Table(df)
    try:
        filtered, _ = MyColSpec.filter(tbl)
        assert expected_columns is not None
        assert set(c.name for c in filtered) == set(expected_columns)
    except ValidationError:
        assert expected_columns is None
    except Exception as e:
        raise AssertionError() from e


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    ("schema", "cast", "success"),
    [
        ({"a": pl.Int64, "b": pl.Int64}, False, False),
        ({"a": pl.String, "b": pl.String}, True, True),
    ],
)
def test_filter_dtypes(schema: dict[str, DataTypeClass], cast: bool, success: bool):
    df = pl.DataFrame(schema=schema)
    try:
        MyColSpec.filter_polars(df, cast=cast)
        assert success
    except DtypeValidationError:
        assert not success
    except Exception as e:
        raise AssertionError() from e
    tbl = pdt.Table(df)
    try:
        MyColSpec.filter(tbl, cast=cast)
        assert success
    except DtypeValidationError:
        assert not success
    except Exception as e:
        raise AssertionError() from e


def fix(counts: dict[str, int]) -> dict[str, int]:
    """Fix the counts to be compatible with the expected output."""
    ret = {k.replace("_primary_key_", "primary_key"): v for k, v in counts.items()}
    return {k: v for k, v in sorted(ret.items(), key=lambda item: item[0])}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
@pytest.mark.parametrize(
    ("data_a", "data_b", "failure_mask", "counts", "cooccurrence_counts"),
    [
        ([1, 2, 3], ["foo", "bar", None], [True, True, True], {}, {}),
        (
            [1, 2, 3],
            ["foo", "bar", "foobar"],
            [True, True, False],
            {"b|max_length": 1},
            {frozenset({"b|max_length"}): 1},
        ),
        (
            [1, 2, 2],
            ["foo", "bar", "foobar"],
            [True, False, False],
            {"b|max_length": 1, "primary_key": 2},
            {
                frozenset({"b|max_length", "primary_key"}): 1,
                frozenset({"primary_key"}): 1,
            },
        ),
    ],
)
def test_filter_failure(
    df_type: type[pl.DataFrame] | type[pl.LazyFrame],
    data_a: list[int],
    data_b: list[str | None],
    failure_mask: list[bool],
    counts: dict[str, int],
    cooccurrence_counts: dict[frozenset[str], int],
):
    df = df_type({"a": data_a, "b": data_b})
    df_valid, failures = MyColSpec.filter_polars(df)
    assert isinstance(df_valid, pl.DataFrame)
    assert_frame_equal(df.filter(pl.Series(failure_mask)).lazy().collect(), df_valid)
    assert validation_mask(df, failures).to_list() == failure_mask
    assert len(failures) == (len(failure_mask) - sum(failure_mask))
    assert failures.counts() == counts
    assert failures.cooccurrence_counts() == cooccurrence_counts
    tbl = pdt.Table(df)
    df_valid, failures = MyColSpec.filter(tbl)
    assert isinstance(df_valid, pdt.Table)
    assert_frame_equal(
        df.filter(pl.Series(failure_mask)).lazy().collect(),
        df_valid >> pdt.export(pdt.Polars()),
    )
    # assert validation_mask(df, failures).to_list() == failure_mask
    # assert len(failures) == (len(failure_mask) - sum(failure_mask))
    assert fix(failures.counts()) == fix(counts)
    # assert failures.cooccurrence_counts() == cooccurrence_counts


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_filter_no_rules(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    class TestColSpec(cs.ColSpec):
        a = cs.Int64(nullable=True)

    df = df_type({"a": [1, 2, 3]})
    df_valid, failures = TestColSpec.filter_polars(df)
    assert isinstance(df_valid, pl.DataFrame)
    assert_frame_equal(df.lazy().collect(), df_valid)
    assert len(failures) == 0
    assert failures.counts() == {}
    assert failures.cooccurrence_counts() == {}
    tbl = pdt.Table(df)
    df_valid, failures = TestColSpec.filter(tbl)
    assert isinstance(df_valid, pdt.Table)
    assert_frame_equal(df.lazy().collect(), df_valid >> pdt.export(pdt.Polars()))
    # assert len(failures) == 0
    assert failures.counts() == {}
    # assert failures.cooccurrence_counts() == {}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_filter_with_rule_all_valid(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    class TestColSpec(cs.ColSpec):
        a = cs.String(min_length=3)

    df = df_type({"a": ["foo", "foobar"]})
    df_valid, failures = TestColSpec.filter_polars(df)
    assert isinstance(df_valid, pl.DataFrame)
    assert_frame_equal(df.lazy().collect(), df_valid)
    assert len(failures) == 0
    assert failures.counts() == {}
    assert failures.cooccurrence_counts() == {}
    tbl = pdt.Table(df)
    df_valid, failures = TestColSpec.filter(tbl)
    assert isinstance(df_valid, pdt.Table)
    assert_frame_equal(df.lazy().collect(), df_valid >> pdt.export(pdt.Polars()))
    # assert len(failures) == 0
    assert failures.counts() == {}
    # assert failures.cooccurrence_counts() == {}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_filter_cast(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    data = {
        # validation: [true, true, false, false, false, false]
        "a": ["1", "2", "foo", None, "123x", "9223372036854775808"],
        # validation: [true, false, true, true, false, true]
        "b": [20, 2000, None, 30, 3000, 50],
    }
    df = df_type(data)
    df_valid, failures = MyColSpec.filter_polars(df, cast=True)
    assert isinstance(df_valid, pl.DataFrame)
    assert df_valid.collect_schema().names() == MyColSpec.column_names()
    assert len(failures) == 5
    assert failures.counts() == {
        "a|dtype": 3,
        "a|nullability": 1,
        "b|max_length": 1,
        # NOTE: primary key constraint is violated as failing dtype casts results in
        # multiple null values.
        "primary_key": 1,
    }
    assert failures.cooccurrence_counts() == {
        frozenset({"a|nullability", "primary_key"}): 1,
        frozenset({"b|max_length"}): 1,
        frozenset({"a|dtype"}): 3,
    }
    tbl = pdt.Table(df)
    df_valid, failures = MyColSpec.filter(tbl, cast=True)
    assert isinstance(df_valid, pdt.Table)
    assert [c.name for c in df_valid] == MyColSpec.column_names()
    assert (
        failures.invalid_rows
        >> pdt.summarize(x=pdt.count())
        >> pdt.export(pdt.Scalar())
        == 5
    )
    assert failures.counts() == {
        "a|dtype": 3,
        "a|nullability": 1,
        "b|max_length": 1,
    }
    # assert failures.cooccurrence_counts() == {
    #     frozenset({"a|nullability", "primary_key"}): 1,
    #     frozenset({"b|max_length"}): 1,
    #     frozenset({"a|dtype"}): 3,
    # }


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_nondeterministic_lazyframe():
    n = 10_000
    lf = pl.LazyFrame(
        {
            "a": range(n),
            "b": [random.choice(["foo", "foobar"]) for _ in range(n)],
        }
    ).select(pl.all().shuffle())

    filtered, _ = MyColSpec.filter_polars(lf)
    assert filtered.select(pl.col("b").n_unique()).item() == 1
    tbl = pdt.Table(lf)
    filtered, _ = MyColSpec.filter(tbl)
    assert (
        filtered
        >> pdt.group_by(filtered.b)
        >> pdt.summarize()
        >> pdt.summarize(x=pdt.count())
        >> pdt.export(pdt.Scalar)
        == 1
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from .const import (
    ALL_COLUMN_TYPES,
    COLUMN_TYPES,
    FLOAT_COLUMN_TYPES,
    INTEGER_COLUMN_TYPES,
    SUPERTYPE_COLUMN_TYPES,
)
from .factory import create_collection, create_collection_raw, create_colspec
from .rules import evaluate_rules_polars, rules_from_exprs_polars

__all__ = [
    "ALL_COLUMN_TYPES",
    "COLUMN_TYPES",
    "FLOAT_COLUMN_TYPES",
    "INTEGER_COLUMN_TYPES",
    "SUPERTYPE_COLUMN_TYPES",
    "create_collection",
    "create_collection_raw",
    "create_colspec",
    "evaluate_rules_polars",
    "rules_from_exprs_polars",
]



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from pydiverse.colspec._validation import validate_columns
from pydiverse.colspec.exc import ValidationError
from pydiverse.colspec.optional_dependency import C, pdt, pl


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_success():
    df = pl.DataFrame(schema={k: pl.Int64() for k in ["a", "b"]})
    tbl = pdt.Table(df)
    out_tbl = validate_columns(tbl, expected=["a"])
    assert set(c.name for c in out_tbl) == {"a"}


@pytest.mark.parametrize(
    ("actual", "expected", "error"),
    [
        (["a"], ["a", "b"], r"1 columns are missing: b;"),
        (["c"], ["a", "b"], r"2 columns are missing: a, b;"),
    ],
)
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_failure(actual: list[str], expected: list[str], error: str):
    df = pl.DataFrame(schema={k: pl.Int64() for k in actual})
    tbl = pdt.Table(df)
    with pytest.raises(ValidationError, match=error):
        validate_columns(tbl, expected=expected)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable
from typing import Any, Concatenate, Literal, ParamSpec, TypeVar, overload

from pydiverse.colspec.optional_dependency import pa


class classproperty(property):  # noqa: N801
    """Replacement for the deprecated @classmethod @property decorator combination.

    Usage:
        ```
        @classproperty
        def num_bytes(self) -> int:
            ...
        ```
    """

    def __get__(self, instance: Any, owner: type | None = None, /) -> Any:
        return self.fget(owner) if self.fget is not None else None


T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")


def map_optional(
    fn: Callable[Concatenate[T, P], R],
    value: T | None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> R | None:
    if value is None:
        return None
    return fn(value, *args, **kwargs)


@overload
def first_non_null(
    *values: T | None, allow_null_response: Literal[True]
) -> T | None: ...


@overload
def first_non_null(*values: T | None, default: T) -> T: ...


def first_non_null(
    *values: T | None,
    default: T | None = None,
    allow_null_response: Literal[True] | None = None,
) -> T | None:
    """Returns the first element in a sequence that is not None."""
    for value in values:
        if value is not None:
            return value
    if allow_null_response:
        return None
    return default


def pydiverse_type_opinions(_type: pa.DataType):
    return (
        pa.string()
        if _type == pa.large_string()
        else pa.time64("us")
        if _type == pa.time64("ns")
        else pa.decimal128(35, 10)
        if pa.types.is_decimal(_type)
        else pa.list_(pydiverse_type_opinions(_type.value_type))
        if pa.types.is_large_list(_type)
        else _type
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec.exc import ValidationError
from pydiverse.colspec.optional_dependency import dy, pl


class FirstColSpec(cs.ColSpec):
    a = cs.Float64()


class SecondColSpec(cs.ColSpec):
    a = cs.String()


class Collection(pydiverse.colspec.collection.Collection):
    first: FirstColSpec
    second: SecondColSpec | None


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_cast_valid(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    first = df_type({"a": [3]})
    second = df_type({"a": [1]})
    out = Collection.cast_polars_data({"first": first, "second": second})  # type: ignore
    assert (
        out.first.collect_schema()
        == FirstColSpec.create_empty_polars().collect_schema()
    )
    assert out.second is not None
    assert (
        out.second.collect_schema()
        == SecondColSpec.create_empty_polars().collect_schema()
    )


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_cast_valid_optional(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    first = df_type({"a": [3]})
    out = Collection.cast_polars_data({"first": first})  # type: ignore
    assert (
        out.first.collect_schema()
        == FirstColSpec.create_empty_polars().collect_schema()
    )
    assert out.second is None


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_cast_invalid_members(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    first = df_type({"a": [3]})
    with pytest.raises(ValueError):
        Collection.cast_polars_data({"third": first})  # type: ignore


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_cast_invalid_member_schema_eager():
    first = pl.DataFrame({"b": [3]})
    with pytest.raises(ValidationError):
        Collection.cast_polars_data({"first": first})


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_cast_invalid_member_schema_lazy():
    first = pl.LazyFrame({"b": [3]})
    collection = Collection.cast_polars_data({"first": first})
    with pytest.raises(ValidationError):
        collection.collect_all_polars()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
from pydiverse.colspec.columns._utils import first_non_null


def test_first_non_null_basic():
    assert first_non_null(1, 2, default=3) == 1
    assert first_non_null(None, 2, default=3) == 2
    assert first_non_null(None, None, default=3) == 3


def test_first_non_null_allow_null_response():
    assert first_non_null(None, None, None, allow_null_response=True) is None


def test_first_non_null_with_terminal():
    assert first_non_null(None, None, None, default=42) == 42
    assert first_non_null(None, 3, None, default=42) == 3


def test_first_non_null_mixed_types():
    assert first_non_null(None, "a", default=3) == "a"
    assert first_non_null(None, 0, default="b") == 0  # 0 is a valid non-null value
    assert (
        first_non_null(None, False, default=1) is False
    )  # False is a valid non-null value


def test_first_non_null_with_kwargs():
    assert first_non_null(None, None, allow_null_response=True) is None
    assert first_non_null(None, None, default="fallback") == "fallback"



# Copyright (c) QuantCo and pydiverse contributors 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable

import pydiverse.common as pdc

from ..optional_dependency import ColExpr, sa
from ._base import Column

# ------------------------------------------------------------------------------------ #


class String(Column):
    """A column of strings."""

    def __init__(
        self,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        min_length: int | None = None,
        max_length: int | None = None,
        regex: str | None = None,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
            min_length: The minimum byte-length of string values in this column.
            max_length: The maximum byte-length of string values in this column.
            regex: A regex that the string values in this column must match. If the
                regex does not use start and end anchors (i.e. ``^`` and ``$``), the
                regex must only be _contained_ in the string.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        super().__init__(
            nullable=nullable, primary_key=primary_key, check=check, alias=alias
        )
        self.min_length = min_length
        self.max_length = max_length
        self.regex = regex

    def dtype(self) -> pdc.String:
        return pdc.String()

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        result = super().validation_rules(expr)

        from pydiverse.colspec.optional_dependency import pl

        len_fn = "len_bytes" if isinstance(expr, pl.Expr) else "len"

        if self.min_length is not None:
            result["min_length"] = getattr(expr.str, len_fn)() >= self.min_length
        if self.max_length is not None:
            result["max_length"] = getattr(expr.str, len_fn)() <= self.max_length
        if self.regex is not None:
            result["regex"] = expr.str.contains(self.regex)
        return result

    def sqlalchemy_column(self, name: str, dialect: sa.Dialect) -> sa.Column:
        """Obtain the SQL column specification of this column definition.

        Args:
            name: The name of the column.
            dialect: The SQL dialect for which to generate the column specification.

        Returns:
            The column as specified in :mod:`sqlalchemy`.
        """
        _ = dialect  # may be used in the future
        return sa.Column(
            name,
            sa.VARCHAR
            if self.max_length is None
            else sa.CHAR(self.min_length)
            if self.min_length == self.max_length
            else sa.VARCHAR(self.max_length),
            nullable=self.nullable,
            primary_key=self.primary_key,
            autoincrement=False,
        )
